/* QwikIndent -- indent structured PIC .asm and .c files using SASM and uncrustify.
 * Written for use in ECE 4175   Fall 2006  by David Bauer (TA)
 */
//#define DEBUG
#define VERSION_NUM  "0.91"
#define USE_INDENT
#define INDENT_OPTIONS "-bad -bli0 -ss -i3 -ip4 -fca -npsl -npcs -nut"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <Windows.h>	// For CopyFile and FileExists
#include <Psapi.h>
// #include <atlconv.h>   <-- C++ only!

/* The program seems short/simple enough to not need a separate header file yet. */
int FileExists(const char *cpFileName);
int IndentC(const char *cpFileName, const char *cppParams[], int iParamCount);
int IndentASM(const char *cpFileName, const char *cppParams[], int iParamCount);
int CopyToBackup(const char *cpFileName);
int MoveBackFile(const char *cpFileName, const char *cpToExt, const char *cpFromExt);
void *MyMalloc(size_t size);
char *MyStrdup(const char *cpStr);
char *AllocStrCat(const char *cpStr1, const char *cpStr2);
char *ChangeExtension(const char *cpFileName, const char *cpNewExt);
char *GetProgramDirectory();

int main(int argc, char *argv[]) {
	char *cpFileExtension;
	char *cpFileName = argv[1];

	if (argc < 2) {
		printf("Qwik Indent (version %s)\nWritten by David Bauer, Fall 2006.\n", VERSION_NUM);
		printf("    ** Requires SASM to indent asm files and Uncrustify to indent C files.\n\n");
		printf("Usage:  qwikindent  <filename> [options]\n");
		exit(-1);
	}

	printf("\"%s\"\n", argv[0]);
	GetProgramDirectory();

	if (argv[1] == 0 || strlen(argv[1]) < 1) {
		printf("Assert Error:  parameter not found.\n");
	}

	cpFileExtension = strrchr(argv[1], '.');
	if (cpFileExtension == NULL) {
		//printf("TODO:  Handle case of no file extension given.\n");
		/* The user did not give the name with a file extension. */
		/* Question, should we always require the file extension, or go searching for it? */

		printf("Please give the full path with extension.\n");
		exit(-2);
	}

	if ( ! FileExists(cpFileName) ) {
		printf("404:  File \"%s\" not found.\n", cpFileName);
		exit(-2);
	}

	if (stricmp(cpFileExtension, ".asm") == 0) {
		IndentASM(argv[1], argv+2, argc-2);
	} else {
		if (stricmp(cpFileExtension, ".c") == 0) {
			IndentC(argv[1], argv+2, argc-2);
		} else {
			printf("File extension unrecognized.  File should either be .c or .asm\n");
		}
	}

	if (cpFileName != argv[1])  free(cpFileName);

	return 0;
}

/* Indent a structured PIC assembly file.
 * Calls on the SASM program (preferrably version 2006.10.08 or later).
 * Passes along extra command-line parameters to SASM.
 */
int IndentASM(const char *cpFileName, const char *cppParams[], int iParamCount) {
	char *cpBuf;
	const char *cpCommand = "sasm -noasm -tabs ";
	char *cpOutFileName;
	int iSize = strlen(cpCommand);
	int i;

	assert(cpFileName != NULL);
	assert(strlen(cpFileName) > 4);
	assert(iParamCount >= 0);

	/* Actions:
	 *   Check that file exists - Done in main line
	 *   Copy file to backup
	 *	 Run SASM on file
	 *   Check output
	 *   Copy .apr over .asm
	 */
	CopyToBackup(cpFileName);

	/* Find the size of the parameters */
	for (i = 0; i < iParamCount; i++) {
		iSize += strlen(cppParams[i]) + 1;
	}
	iSize += strlen(cpFileName);

	/* And allocate buffer space. */
	cpBuf = (char *) MyMalloc(sizeof(char) * (iSize + 4));

	strcpy(cpBuf, cpCommand);
	for (i = 0; i < iParamCount; i++) {
		strcat(cpBuf, cppParams[i]);
		strcat(cpBuf, " ");
	}
	strcat(cpBuf, cpFileName);

#ifdef DEBUG
	printf("Debug:  Command line: \"%s\"\n", cpBuf);
#endif

	system(cpBuf);
	free(cpBuf);

	cpOutFileName = ChangeExtension(cpFileName, ".apr");	// Get the output file name
	if (cpOutFileName == NULL) return -1;

	if (FileExists(cpOutFileName)) {					// Check to see if it exists (i.e., the indenting worked)
		MoveBackFile(cpFileName, ".asm", ".apr");		// If so, copy it over the input
	} else {
		printf("Error:  Operation failed (no output file created).\n");	// Else, print error!
		free(cpOutFileName);
		return -1;
	}
	free(cpOutFileName);

	return 0;
}

/* Indent a c file.
 * Calls on the uncrustify program.  Requires the configuration file (4175.cfg) to be in the expected place.
 * Passes along extra command-line parameters to uncrustify.
 */
int IndentC(const char *cpFileName, const char *cppParams[], int iParamCount) {
	char *cpBuf;
	char *cpOutFileName;
	char *cpBaseCFGPath;
	char cpCommand[256 + 24];
	int iSize = sizeof(cpCommand);
	int i;

	assert(cpFileName != NULL);
	assert(strlen(cpFileName) > 2);
	assert(iParamCount >= 0);
	assert(iSize == (256 + 24));

	if (strlen(cpFileName) > 256) {
		printf("File name is too long!\n");
		return -1;
	}

#ifndef USE_INDENT
	cpBaseCFGPath = GetProgramDirectory();
	if (cpBaseCFGPath == NULL) {
		printf("Error:  Failed to get program path.\n");
		return -1;
	}

	sprintf(cpCommand, "uncrustify -c %s4175.cfg ", cpBaseCFGPath);
#else
	sprintf(cpCommand, "indent %s ", INDENT_OPTIONS);
#endif

	/* Actions:
	 *    Check that file exists - Done in main line
	 *    Copy file to backup
	 *    Run uncrustify on file
	 *    Check output
	 *    Copy .uncrustify over .c
	 */
	CopyToBackup(cpFileName);

	/* Find the size of the parameters */
	for (i = 0; i < iParamCount; i++) {
		iSize += strlen(cppParams[i]) + 1;
	}
	iSize += strlen(cpFileName);

	/* And allocate buffer space. */
	cpBuf = (char *) MyMalloc(sizeof(char) * (iSize + 4));

	strcpy(cpBuf, cpCommand);
	for (i = 0; i < iParamCount; i++) {
		strcat(cpBuf, cppParams[i]);
		strcat(cpBuf, " ");
	}
	strcat(cpBuf, cpFileName);

#ifdef DEBUG
	printf("Debug:  Command line: \"%s\"\n", cpBuf);
#endif


	system(cpBuf);	/* Call the program to perform the actual indentation. */
	free(cpBuf);

#ifndef USE_INDENT
	cpOutFileName = ChangeExtension(cpFileName, ".c.uncrustify");	// Get the output file name
	if (cpOutFileName == NULL) return -1;

	if (FileExists(cpOutFileName)) {					// Check to see if it exists (i.e., the indenting worked)
		MoveBackFile(cpFileName, ".c", ".c.uncrustify");	// .c.uncrustify -> .c
	} else {
		printf("Error:  Operation failed (no output file created).\n");	// Else, print error!
		free(cpOutFileName);
		return -1;
	}
	free(cpOutFileName);
#endif

	return 0;
}

/* Checks to see if the specified file exists and can be opened for reading.
 * Returns zero if not, and non-zero if it can.
 */
int FileExists(const char *cpFileName) {
	FILE *fp;

	if (NULL != (fp = fopen(cpFileName, "rb"))) {
		fclose(fp);
		return 1;
	}

	return 0;
}

/* Returns a new string containing the base name of the given file name.
 * MAY RETURN THE ORIGINAL POINTER -- This should probably be changed for consistency.
 */
char *BaseFileName(const char *cpFileName) {
	char *cpBuf;
	char *cpPeriod;

	assert(cpFileName != NULL);

	cpBuf = MyStrdup(cpFileName);

	cpPeriod = strrchr(cpBuf, '.');
	if (cpPeriod == NULL) {
		free(cpBuf);
		return (char *) cpFileName;		/* Get rid of warning. */
	}
	*cpPeriod = 0;	/* Easy way to cut off the file extension. */

	return cpBuf;
}

/* Wrapper around malloc that checks that the allocation succeeded. */
void *MyMalloc(size_t size) {
	void *vp = malloc(size);

	if (vp == NULL) {
		fprintf(stderr, "Malloc: Out of memory!\n");
		exit(-10);
	}

	return vp;
}

/* Wrapper around strdup that checks that the allocation succeeded. */
char *MyStrdup(const char *cpStr) {
	char *cpOut = strdup(cpStr);

	if (cpOut == NULL) {
		fprintf(stderr, "Strdup: Out of memory!\n");
		exit(-11);
	}

	return cpOut;
}

/* Copy the given file (named "file.ext") to a numbered backup file.
 * Backup files are named "file.000", "file.001", etc.
 */
int CopyToBackup(const char *cpFileName) {
	int i = 0;
	char *cpBuf;
	char *cpExt;

	assert(cpFileName != NULL);

	cpExt = strrchr(cpFileName, '.');
	assert(cpExt != NULL);

	cpBuf = (char *) MyMalloc(sizeof(char)*(strlen(cpFileName) + 4));
	
	strncpy(cpBuf, cpFileName, cpExt - cpFileName);		/* Copy up until the period. */
	
	cpExt = cpBuf + (cpExt - cpFileName);				/* Set to point to the extension */

	/* Check for already existing backup files and find the next free number for the extension. */
	do {
		sprintf(cpExt, ".%03d", i++);
	} while (FileExists(cpBuf) && i < 999);

	if (i > 5) {
		printf("Warning:  there are a lot of backup files (.000, .001, etc)\n");
	}

	if (FileExists(cpBuf)) {
		printf("Warning:  Overwriting file \"%s\"!  Please delete extra backup files (.000, .001, etc)!\n", cpBuf);
	}

	printf("Backup file: \"%s\"\n", cpBuf);
	
	CopyFile(cpFileName, cpBuf, 1);

	return 0;
}

/* Remember:  FileName = "file.asm", ToExt = ".asm", FromExt = ".apr"
 *            Filename = "file.c", ToExt = ".c", FromExt = ".c.uncrustify" */
int MoveBackFile(const char *cpFileName, const char *cpToExt, const char *cpFromExt) {
	char *cpBaseName = BaseFileName(cpFileName);
	char *cpBuf;
	const char *cpExt;

	assert(cpBaseName != NULL);
	assert(cpToExt != NULL);
	assert(cpFromExt != NULL);

	cpExt = strrchr(cpFileName, '.');

	if (stricmp(cpExt, cpToExt)) {
		fprintf(stderr, "Error:  file extension does not match expected value (\"%s\" != \"%s\")\n", cpExt, cpToExt);
		exit(-14);
	}

	cpBuf = (char *) MyMalloc(sizeof(char) * (strlen(cpBaseName) + strlen(cpFromExt) + 4));
	if (cpBuf == NULL) return -1;
	strcpy(cpBuf, cpBaseName);
	strcat(cpBuf, cpFromExt);

#ifdef DEBUG
	printf("Debug:  Copying \"%s\" to \"%s\"\n", cpBuf, cpFileName);
#endif
	if (!FileExists(cpBuf)) {
		printf("Assert error:  Source file does not exist in MoveBackFile\n");
		exit(-15);
	}
	DeleteFile(cpFileName);
	MoveFile(cpBuf, cpFileName);

	if (cpBaseName != cpFileName) free(cpBaseName);
	return 0;
}

/* Returns a new string (allocated by malloc) containing the concatenation of the given
 * strings.  Does not modify the given strings.
 */
char *AllocStrCat(const char *cpStr1, const char *cpStr2) {
	char *cpBuf;

	assert(cpStr1 != NULL);
	assert(cpStr2 != NULL);

	cpBuf = (char *) MyMalloc(sizeof(char) * (strlen(cpStr1) + strlen(cpStr2) + 1));
	if (cpBuf == NULL) return NULL;

	strcpy(cpBuf, cpStr1);
	strcat(cpBuf, cpStr2);

	return cpBuf;
}

/* Returns a string (allocated by malloc) containing the given string,
 * with the extension (after the final '.') replaced by the new extension.
 */
char *ChangeExtension(const char *cpFileName, const char *cpNewExt) {
	char *cpBaseName = BaseFileName(cpFileName);
	char *cpBuf;

	assert(cpBaseName != NULL);
	assert(cpFileName != NULL);
	assert(cpNewExt != NULL);
	
	cpBuf = AllocStrCat(cpBaseName, cpNewExt);

	free(cpBaseName);
	return (cpBuf);
}

/* Get program directory as an ASCII string. (Returned string is allocated by malloc.)
 * Thanks to Louis for finding GetModuleFileName, which is much cleaner than GetModuleFileNameEx.
 */
char *GetProgramDirectory() {
	char cpBuf[512];		// 256 * 2 because I don't know whether Unicode is defined or not
	char *cpDirPath;
	char *cpLastSlash;
	LPTSTR lpFilename = cpBuf;

	GetModuleFileName(NULL, lpFilename, 256);
	
	//cpDirPath = strdup(T2A(lpFilename));		/* Convert from Unicode if necessary */
	cpDirPath = strdup(lpFilename);
	cpLastSlash = strrchr(cpDirPath, '\\');		/* Assume DOS slash convention */
	if (cpLastSlash == NULL) return NULL;

	*(cpLastSlash + 1) = 0;						/* Truncate the string at the final slash */

	return cpDirPath;
}
